class BaseError(Exception):
    """自定义异常基类，通用抽象接口"""

    def __init__(self, message: str = ""):
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return self.message if self.message else "BaseError"


class ImageGetError(BaseError):
    """自定义异常类，用于表示图片获取失败的情况。"""

    link: str
    e: Exception

    def __init__(
        self, link: str = "", message: str = "图片获取失败", e: Exception = None
    ):
        self.link = link
        self.e = e
        super().__init__(message)

    def __str__(self) -> str:
        return f"{self.message}. link={self.link}. {self.e}"


class NOException(BaseError):
    """没有异常"""

    pass



import asyncio
import time

from httpx import AsyncClient


class AutoCloseAsyncClient(AsyncClient):
    """
    1.自动关闭
    AutoCloseAsyncClient会在指定的超时（timeout）后自动关闭，如果在这段时间内没有调用任何方法。
    2.上下文管理
    使用异步上下文管理器（__aenter__和__aexit__），在使用完客户端后确保资源得到清理。
    """

    # 超时时间，默认为30s
    auto_close_time: int
    last_action_time = time.monotonic()
    _running = True

    def __init__(self, auto_close_time: int = 30, *args, **kwargs):
        self.auto_close_time = auto_close_time
        self._timeout_task = asyncio.create_task(self._check_timeout())
        kwargs["trust_env"] = False  # ban proxy
        super().__init__(*args, **kwargs)

    async def _check_timeout(self):
        while self._running:
            await asyncio.sleep(1)
            if time.monotonic() - self.last_action_time > self.auto_close_time:
                await self.close()

    async def _update_action_time(self):
        self.last_action_time = time.monotonic()

    async def get(self, *args, **kwargs):
        await self._update_action_time()
        return await super().get(*args, **kwargs)

    async def post(self, *args, **kwargs):
        await self._update_action_time()
        return await super().post(*args, **kwargs)

    async def close(self):
        self._running = False
        await super().aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()
        self._timeout_task.cancel()


# 使用示例
async def main():
    async with AsyncClient(trust_env=False) as client:
        print(await client.get("https://baidu.com"))  # 发起请求

    client = AutoCloseAsyncClient(auto_close_time=3)
    print(await client.get("https://baidu.com"))  # 发起请求

    await asyncio.sleep(1)  # 等待小于超时的时间

    print("Waiting for timeout...")
    await asyncio.sleep(3)  # 等待大于超时的时间
    print("Client should be closed now.")


# 运行示例
if __name__ == "__main__":
    asyncio.run(main())



"""class tools"""

import copy
import time


def copy_attr(
    o: object,
    other: object,
    add_new_attr: bool = True,
    cover: bool = True,
):
    """copy o's attr to other, by deepcopy.

    Args:
        o (object): copy from: o
        other (object): copy to: other
        add_new_attr (bool, optional):whether to add new attr,
        if `other` has not attr. Defaults to True.
        cover (bool, optional): whether to cover the value of `other.attr`,
        if `other.attr` is not None. Defaults to True.
    """
    b1 = add_new_attr
    b3 = cover
    for k, v in vars(o).items():
        b2 = hasattr(other, k)
        b4 = getattr(other, k) is None
        # 2. bool calculate
        if (b1 and ((not b2) or b4)) or (b2 and (b3 or b4)):
            setattr(other, k, copy.deepcopy(v))
        # 1.最普通的方法
        # if b1 and not b2 or b4:
        #     # 如果允许添加新属性 (条件1), 并且没有旧的属性(条件2）或者没有有效的属性（条件4）
        #     setattr(other, k, copy.deepcopy(v))
        # elif b2:  # 如果目标对象含有该属性(条件2)，才可能会赋值
        #     if b3:  # 如果允许覆盖(条件3)，直接拷贝
        #         setattr(other, k, copy.deepcopy(v))
        #     elif b4:  # 如果不允许覆盖，当other.attr为None时(条件4)候拷贝
        #         setattr(other, k, copy.deepcopy(v))


if __name__ == "__main__":

    class T:
        x: int
        y: int

    start = time.time()
    for i in range(1234567):
        a = T()
        a.x = 1
        a.y = 2
        b = T()
        b.x = 1
        b.y = None
        copy_attr(a, b)
    end = time.time()
    print(end - start)



import inspect
from inspect import Parameter
from collections.abc import Callable
from functools import wraps


def compat_arg_error(func: Callable = None, rm_None: bool = True) -> Callable:
    """这个装饰器使得函数在接收错误的参数时，仍然正常执行，消除异常。（但不保证完全正确）
    正常情况下，在调用函数时传入了错误参数，Python 会抛出 TypeError 异常。有2种情况：
    1.位置参数：如果你传入了超过函数定义的参数数量，或者传入了一个不接受的位置参数，Python 会报错。
    2.关键字参数：如果你传入了不在函数定义中的关键字参数，Python 也会抛出 TypeError。

    Args:
        func (Callable): 被装饰的函数
        rm_None(bool) : 是否移除为None的参数. (default True)
    Returns:
        Callable: 装饰后的函数
    """
    if func is None:
        return lambda f: compat_arg_error(f)

    @wraps(func)
    def wrapper(*args, **kwargs):
        signature = inspect.signature(func)
        params = signature.parameters
        # 1.过滤掉未定义的关键字参数, 以及None(当rm_None为True)
        filtered_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k in params
            and not (rm_None and v is None)
            and (
                params[k].kind
                in {
                    Parameter.KEYWORD_ONLY,
                    Parameter.VAR_KEYWORD,
                    Parameter.POSITIONAL_OR_KEYWORD,
                }
            )
        }
        # 2.位置参数
        # 2.1.类型检查，自动类型转换
        for i, (k, v) in enumerate(params.items()):
            if i < len(args):
                true_type = v.annotation
                if true_type is not Parameter.empty:
                    try:
                        args[i] = true_type(args[i])  # 尝试转换
                    except (TypeError, ValueError):
                        raise TypeError.with_traceback(
                            f"Warning: arg '{k}' need type{true_type}"
                        )

        # 2.2.数量检查，减少多余参数
        position_param_nums = sum(
            1
            for p in params.values()
            if p.kind
            in {
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            }
        )
        if len(args) > position_param_nums:
            args = args[:position_param_nums]

        # 3.正确性检验
        try:
            bound_args = signature.bind_partial(*args, **filtered_kwargs)
            bound_args.apply_defaults()  # 应用默认值
        except TypeError as e:
            raise TypeError(f"TypeError! Bug detected! Please report this issue:\n{e}")
        return func(*bound_args.args, **filtered_kwargs)

    return wrapper


if __name__ == "__main__":

    # 示例用法
    @compat_arg_error
    def example_function(a, b, c=10):
        return a + b + c

    result = example_function("6", 1, 2, "9", 9, d="")  # d 参数将被忽略
    print(result)  # 输出

    # 示例用法2
    @compat_arg_error()
    def example_function2(a, b, c=10):
        return a + b + c

    # 测试
    result2 = example_function2("", 1, 2, "9", 9, d="")  # d 参数将被忽略
    print(result2)  # 输出



import asyncio
from pathlib import Path
from urllib.request import urlopen

from PIL import Image
from io import BytesIO

from naotool.exception import ImageGetError
from naotool.httpn import AutoCloseAsyncClient


# 最外层接口get，大一统接口
async def get(links: Path | str | list, client=None) -> Image.Image | list[Image.Image]:
    """下载单张图片、或批量图片

    Args:
        links (Path | str | list): 链接列表, 或者链接str, Path
        client (AsyncClient, None): 用户可以自定义client用于代理、反爬等等, 默认不使用代理.

    Returns:
        Image.Image | list[Image.Image]: Image图片,or图片列表
    """
    # 1.统一处理为字符串列表
    one = not isinstance(links, list)
    if isinstance(links, str):
        links = [links]
    if isinstance(links, Path):
        links = [str(links)]
    # 2.处理client
    async with AutoCloseAsyncClient() if not client else client as client:
        imgs = await get_imgs(links, client)
    return imgs[0] if one and imgs else imgs


async def get_local(link: str) -> Image.Image:
    """打开本地图片"""
    if not link.startswith("file:"):
        raise ImageGetError(link, f"\nget_local error.")
    with urlopen(link) as response:
        file_data = response.read()
        return Image.open(BytesIO(file_data)).convert("RGB")


async def get_http(link: str, client: AutoCloseAsyncClient) -> Image.Image:
    """下载网络图片"""
    if not link.startswith("http"):
        raise ImageGetError(link, f"\nget_local error.")
    res = await client.get(link)
    if res.is_error:
        raise ImageGetError(link, f"\nget_http error.")
    return Image.open(BytesIO(res.content)).convert("RGB")


async def get_path(link: str) -> Image.Image:
    return Image.open(Path(link)).convert("RGB")


async def get_img(link: str, client: AutoCloseAsyncClient) -> Image.Image:
    """get一个图片"""
    try:
        if link.startswith("http"):
            return await get_http(link, client)
        elif link.startswith("file"):
            return await get_local(link)
        else:
            return await get_path(link)
    except ImageGetError:
        raise
    except Exception as e:
        raise ImageGetError(link, e=e)


async def get_imgs(links: list[str], client: AutoCloseAsyncClient) -> list[Image.Image]:
    """批量下载图片"""
    img_list = []
    # 并行
    tasks = (get_img(link, client) for link in links)
    res_list = await asyncio.gather(*tasks)
    img_list.extend(filter(None, res_list))
    return img_list


async def test():
    images = await get(
        [
            "https://avatars.githubusercontent.com/u/53679884",
            "https://avatars.githubusercontent.com/u/53679884",
        ]
    )
    images[0].save("tmp.jpg", format="JPEG")


if __name__ == "__main__":
    images = asyncio.run(test())



import hashlib
from PIL import Image


def img_md5hex(img: Image.Image) -> str:
    """计算一个图片的md5"""
    # 将图片转换为二进制数据
    img_bytes = img.tobytes()
    # 创建一个MD5哈希对象
    md5_hash = hashlib.md5()
    # 更新哈希对象，img一般不会大，不用分块处理
    md5_hash.update(img_bytes)
    return md5_hash.hexdigest()



from typing import Callable


def tight(
    s1: list[str] | str, sep1: str = "\n", sep2: str = " ", ext_func: Callable = None
) -> str:
    """去除大部分空白字符，自定义1、2级分隔符

    Args:
        s_list (list[str]): 原始str list
        sepa1 (str, optional): 1级别分隔符, 用于分隔拼接原始字符串列表. Defaults to "\\n".
        sepa2 (str, optional): 2级别分隔符, 用于分隔拼接原始字符串。Defaults to " ".
        ext_func (Callable, optional): 自定义过滤函数。Defaults to `None`
    Returns:
        str:
    """
    if isinstance(s1, list):
        return sep1.join(
            ext_func(
                sep2.join(ext_func(s3.strip()) for s3 in s2.split(sep2) if s3.strip())
            )
            for s2 in s1
            if s2.strip()
        )
    elif isinstance(s1, str):
        return sep1.join(
            ext_func(
                sep2.join(ext_func(s3.strip()) for s3 in s2.split(sep2) if s3.strip())
            )
            for s2 in s1.split(sep1)
            if s2.strip()
        )
    raise TypeError(s1, "only support type str, list[str]")


if __name__ == "__main__":
    s = """   去除大部分
    空     白    字        符，
    自定义1、2级分隔符
    """
    s = tight(s)
    print(s)
